// version:0.03
/++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 + Program Name: hilevel
 + Coded by Larry Cowan
 + 29 February 2004
 +
 + This program will read one or more D language modules and display all
 +     information which is in global scope.  // Comments are preserved,
 +     but /star and /plus comments are ignored.
 + All printed data is prefixed by line number.
 + Versioned areas are separated by empty lines.
 + Warning: non-global versioning is ignored.
 +
 + '-d' parms can be inserted in the command line to flipflop printing
 +  of additional diagnostic output.
 + This is forced permanently on if compiled in debug (hilevel) mode.
 +
 + ToDo: I should allow user to set the depth desired, to adapt to different
 +       types of input and coding styles.  Maybe later...
 +
 + Wildcard naming is supported for the arguments, with everything being
 + case-sensitive for unix and case-insensitive for win32.
 +
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 +  Copyright 2004 by Larry F. Cowan
 +  Distribution rights:  This code may be used for any purpose whatever,
 +      commercial or otherwise, without specific permission from the 
 +      author(s), as long as they are not misrepresented in any way, the
 +      author name(s) are kept in the source, and this notice is kept.
 +      Changes should indicate at least an author name and date.  No listed
 +      author takes on any responsibility or liability for your use of this
 +      module. It's free.  Caveat emptor!  The underlying (imported) modules
 +      are under their own licensing and this is not affected here. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++/
//debug=hilevel
version=win32;

import std.c.stdio;
import streams;
import expandw;

void main ( char[][] args )                 // PROGRAM MAIN ENTRY POINT
{
    WildCard wc = new WildCard();
    bit diag = false;
    
    if (args.length < 2) {                  // err: no args
        printf("Usage: hilevel file1.d [file2.d [file3.d [...]]]\n");
    }
    args = args[1..args.length];            // drop (arg0) progname
    foreach ( char[] arg; args ) {          // for each argument
        if (arg == "-d") {
            diag = (diag) ? false : true;   // flipflop diag switch
            continue;
        }
        foreach ( char[] fname; wc.expand(arg) ) {
            try {
                ReadableFile inp = new ReadableFile(fname); // open file stream
                processFile(inp,diag);                      // display it
                inp.close();                                // close it
            }
            catch (Error e) {  // opening or processing failed - show error
                  printf("\n*Error* : %.*s\n",e.msg);
            }
        }
    } // and see if we have another filearg or switch
}

/*******************************************************************
 * The following is a hand-coded state machine to select and print
 * the appropriate data from an input file stream.
 */
void processFile ( ReadableFile inp, bit diag )
{
    char[] buf;
    int lineno = 0;
    int inbraces = 0;
    bit version_beg = false;
    bit version_on = false;
    bit version_end = false;
    bit apos = false;
    bit quot = false;
    bit tic = false;
    bit comment = false;
    bit cutout = false;
    bit printit = false;
    bit newline = false;

    debug (hilevel) { diag = true; }  // forces constant diags on

    printf("
*********************** %.*s **********************\n\n",inp.filename());

    while (1) {
        buf = inp.readLine();    // fill line buffer
        if (inp.eof())           // no more data
            break;
        lineno++;                // new line - count it
        newline = false;         // new line - no printable data yet
        tic = false;             // no carryover of dualchar starts

        if (version_end) {
            putchar('\n');       // blankline after versioning closes w/'}'
            version_end = false;
        }
        if (diag) {              // intersperse diagnostics
            if (inbraces)
                printf("%04d: inbraces = %d\n",lineno,inbraces);
            else
            if (cutout || comment || quot || apos)
                printf("%04d: cutout=%d, comment=%d, quot=%d, apos=%d\n"
                      , lineno
                      , cutout
                      , comment
                      , quot
                      , apos );
        }
        if (buf.length == 0 || (buf[0] == '\n' || buf[0] == '\r'))
            continue;                       // drop empty lines immediately
        if (buf.length > 1 && buf[0..2] == "//") {
            printf("%4d: %.*s\n",lineno,buf); // show comment, go to next line
            continue;
        }
        if (inbraces == 0 && buf.length > 7 && buf[0..7] == "version") {
            version_beg = true; // versioning at global level stays global
            printf("\n");
        }
        int chn = 0;
        foreach ( char ch; buf ) {
            debug (hilevel) { printf("[%c.%c.%d]\n",(tic)?'+':'-',ch,chn++); }
            if (ch == '}' && !cutout && !comment && !quot && !apos) {
                if (version_on && inbraces == 0) {
                    version_on = false;
                    version_end = true;
                }
                else
                    inbraces--;
                if (inbraces < 0) // a little protection
                    inbraces = 0;
                tic = false;
                continue;
            }
            if (ch == '{' && !cutout && !comment && !quot && !apos) {
                if (version_beg) {
                    version_beg = false;
                    version_on = true;
                } else
                    inbraces++;
                tic = false;
                continue;
            }
            if (inbraces == 0 && !cutout && !comment) { // in global scope?
                if (ch != '\n' && ch != '\r') { // ignore end-of-line-chars
                    if (!newline)
                        printf("%4d: ",lineno); // first time, start new line
                    putchar(ch);                // print current character
                    newline = true;             // note line has accepted data
                }
            }
            if (inbraces > 0) {
                tic = false;
                continue;
            }
            if (ch == '\'' && !cutout && !comment) { // apos outside of comment
                if (!quot)
                    apos = (apos) ? false : true; // (and apos) should be paired
                tic = false;
                continue;
            }
            if (ch == '"' && !cutout && !comment) { // quot outside of comment
                if (!apos)
                    quot = (quot) ? false : true; // (and quot) should be paired
                tic = false;
                continue;
            }
            if (ch == '/') {
                if (!cutout && !comment) {  // found '/' outside of comment
                    if (tic) {
                        tic = false;
                        break;         // found '//' ignore rest of line
                    }
                    tic = true;        // mark as slash pending
                    continue;
                }
                if (tic && cutout)   
                    cutout = false;    // found '+/'
                if (tic && comment)
                    comment = false;   // found '+/'
                tic = false;
                continue;
            }
            if (ch == '+' && !comment) {
                if (tic) {             // found '+' of '/+'
                    cutout = true;     // mark as in-a-/+-comment
                    tic = false;
                    continue;
                }
                if (cutout) {          // was looking for '+/'
                    tic = true;        // mark as '+' pending  
                    continue;
                }
            }
            if (ch == '*' && !cutout) {
                if (tic) {             // found '*' of '/*'
                    comment = true;    // mark as in-a-/*-comment
                    tic = false;
                    continue;
                }
                if (comment) {         // was looking for '*/'
                    tic = true;        // mark as '*' pending
                    continue;
                }
            }
            tic = false;               // unused tic dies
        }
        if (newline)                   // if line was started, end it
            putchar('\n');
    }
    return;
}

// end-of-program
